@@ -1,4 +1,4 @@
-#!/bin/bash
+#!/bin/sh
 #
 # Licensed to the Apache Software Foundation (ASF) under one
 # or more contributor license agreements.  See the NOTICE file
@@ -18,8 +18,7 @@
 #
 
 BASEDIR=`dirname $0`
-BASE_BRANCH=cassandra-4.0
-set -e
+BASE_BRANCH=cassandra-4.1
 
 die ()
 {
@@ -30,17 +29,13 @@ die ()
 
 print_help()
 {
-  echo "Usage: $0 [-f|-p|-a|-e|-i|-b|-s]"
-  echo "   -a Generate the config.yml, config.yml.FREE and config.yml.PAID expanded configuration"
-  echo "      files from the main config_template.yml reusable configuration file."
-  echo "      Use this for permanent changes in config that will be committed to the main repo."
-  echo "   -f Generate config.yml for tests compatible with the CircleCI free tier resources"
-  echo "   -p Generate config.yml for tests compatible with the CircleCI paid tier resources"
-  echo "   -b Specify the base git branch for comparison when determining changed tests to"
-  echo "      repeat. Defaults to ${BASE_BRANCH}. Note that this option is not used when"
-  echo "      the '-a' option is specified."
-  echo "   -s Skip automatic detection of changed tests. Useful when you need to repeat a few ones,"
-  echo "      or when there are too many changed tests for CircleCI."
+  echo "Usage: $0 [-l|-m|-h|-f|-e]"
+  echo "   -a Generate the default config.yml using low resources and the three templates"
+  echo "      (config.yml.LOWRES, config.yml.MIDRES and config.yml.HIGHRES). Use this for"
+  echo "      permanent changes in config-2_1.yml that will be committed to the main repo."
+  echo "   -l Generate config.yml using low resources"
+  echo "   -m Generate config.yml using mid resources"
+  echo "   -h Generate config.yml using high resources"
   echo "   -e <key=value> Environment variables to be used in the generated config.yml, e.g.:"
   echo "                   -e DTEST_BRANCH=CASSANDRA-8272"
   echo "                   -e DTEST_REPO=https://github.com/adelapena/cassandra-dtest.git"
@@ -53,55 +48,55 @@ print_help()
   echo "                   -e REPEATED_UTESTS_LONG_COUNT=100"
   echo "                   -e REPEATED_UTESTS_STRESS=org.apache.cassandra.stress.generate.DistributionGaussianTest"
   echo "                   -e REPEATED_UTESTS_STRESS_COUNT=500"
+  echo "                   -e REPEATED_SIMULATOR_DTESTS=org.apache.cassandra.simulator.test.TrivialSimulationTest"
+  echo "                   -e REPEATED_SIMULATOR_DTESTS_COUNT=500"
   echo "                   -e REPEATED_JVM_DTESTS=org.apache.cassandra.distributed.test.PagingTest"
   echo "                   -e REPEATED_JVM_DTESTS_COUNT=500"
   echo "                   -e REPEATED_JVM_UPGRADE_DTESTS=org.apache.cassandra.distributed.upgrade.GroupByTest"
   echo "                   -e REPEATED_JVM_UPGRADE_DTESTS_COUNT=500"
   echo "                   -e REPEATED_DTESTS=cdc_test.py cqlsh_tests/test_cqlsh.py::TestCqlshSmoke"
   echo "                   -e REPEATED_DTESTS_COUNT=500"
-  echo "                   -e REPEATED_LARGE_DTESTS=replace_address_test.py::TestReplaceAddress::test_replace_stopped_node"
+  echo "                   -e REPEATED_LARGE_DTESTS=consistency_test.py::TestAvailability::test_network_topology_strategy"
   echo "                   -e REPEATED_LARGE_DTESTS=100"
   echo "                   -e REPEATED_UPGRADE_DTESTS=upgrade_tests/cql_tests.py upgrade_tests/paging_test.py"
   echo "                   -e REPEATED_UPGRADE_DTESTS_COUNT=25"
   echo "                   -e REPEATED_ANT_TEST_TARGET=testsome"
   echo "                   -e REPEATED_ANT_TEST_CLASS=org.apache.cassandra.cql3.ViewTest"
   echo "                   -e REPEATED_ANT_TEST_METHODS=testCompoundPartitionKey,testStaticTable"
+  echo "                   -e REPEATED_ANT_TEST_VNODES=false"
   echo "                   -e REPEATED_ANT_TEST_COUNT=500"
   echo "                  For the complete list of environment variables, please check the"
-  echo "                  list of examples in config_template.yml and/or the documentation."
+  echo "                  list of examples in config-2_1.yml and/or the documentation."
   echo "                  If you want to specify multiple environment variables simply add"
-  echo "                  multiple -e options. The flags -f/-p should be used when using -e."
-  echo "   -i Ignore unknown environment variables"
+  echo "                  multiple -e options. The flags -l/-m/-h should be used when using -e."
+  echo "   -f Stop checking that the environment variables are known"
 }
 
 all=false
-free=false
-paid=false
+lowres=false
+midres=false
+highres=false
 env_vars=""
 has_env_vars=false
 check_env_vars=true
-detect_changed_tests=true
-while getopts "e:afpib:s" opt; do
+while getopts "e:almhf" opt; do
   case $opt in
       a ) all=true
-          detect_changed_tests=false
           ;;
-      f ) free=true
+      l ) lowres=true
           ;;
-      p ) paid=true
+      m ) midres=true
           ;;
-      e ) if (!($has_env_vars)); then
+      h ) highres=true
+          ;;
+      e ) if (! ($has_env_vars)); then
             env_vars="$OPTARG"
           else
             env_vars="$env_vars|$OPTARG"
           fi
           has_env_vars=true
           ;;
-      b ) BASE_BRANCH="$OPTARG"
-          ;;
-      i ) check_env_vars=false
-          ;;
-      s ) detect_changed_tests=false
+      f ) check_env_vars=false
           ;;
       \?) die "Invalid option: -$OPTARG"
           ;;
@@ -127,6 +122,8 @@ if $has_env_vars && $check_env_vars; then
        [ "$key" != "REPEATED_UTESTS_LONG_COUNT" ] &&
        [ "$key" != "REPEATED_UTESTS_STRESS" ] &&
        [ "$key" != "REPEATED_UTESTS_STRESS_COUNT" ] &&
+       [ "$key" != "REPEATED_SIMULATOR_DTESTS" ] &&
+       [ "$key" != "REPEATED_SIMULATOR_DTESTS_COUNT" ] &&
        [ "$key" != "REPEATED_JVM_DTESTS" ] &&
        [ "$key" != "REPEATED_JVM_DTESTS_COUNT" ] &&
        [ "$key" != "REPEATED_JVM_UPGRADE_DTESTS" ]  &&
@@ -140,62 +137,67 @@ if $has_env_vars && $check_env_vars; then
        [ "$key" != "REPEATED_ANT_TEST_TARGET" ] &&
        [ "$key" != "REPEATED_ANT_TEST_CLASS" ] &&
        [ "$key" != "REPEATED_ANT_TEST_METHODS" ] &&
+       [ "$key" != "REPEATED_ANT_TEST_VNODES" ] &&
        [ "$key" != "REPEATED_ANT_TEST_COUNT" ]; then
       die "Unrecognised environment variable name: $key"
     fi
   done
 fi
 
-if $free; then
-  ($all || $paid) && die "Cannot use option -f with options -a or -p"
-  echo "Generating new config.yml file for free tier from config_template.yml"
-  circleci config process $BASEDIR/config_template.yml > $BASEDIR/config.yml.FREE.tmp
-  cat $BASEDIR/license.yml $BASEDIR/config.yml.FREE.tmp > $BASEDIR/config.yml
-  rm $BASEDIR/config.yml.FREE.tmp
+if $lowres; then
+  ($all || $midres || $highres) && die "Cannot use option -l with options -a, -m or -h"
+  echo "Generating new config.yml file with low resources from config-2_1.yml"
+  circleci config process $BASEDIR/config-2_1.yml > $BASEDIR/config.yml.LOWRES.tmp
+  cat $BASEDIR/license.yml $BASEDIR/config.yml.LOWRES.tmp > $BASEDIR/config.yml
+  rm $BASEDIR/config.yml.LOWRES.tmp
 
-elif $paid; then
-  ($all || $free) && die "Cannot use option -p with options -a or -f"
-  echo "Generating new config.yml file for paid tier from config_template.yml"
-  patch -o $BASEDIR/config_template.yml.PAID $BASEDIR/config_template.yml $BASEDIR/config_template.yml.PAID.patch
-  circleci config process $BASEDIR/config_template.yml.PAID > $BASEDIR/config.yml.PAID.tmp
-  cat $BASEDIR/license.yml $BASEDIR/config.yml.PAID.tmp > $BASEDIR/config.yml
-  rm $BASEDIR/config_template.yml.PAID $BASEDIR/config.yml.PAID.tmp
+elif $midres; then
+  ($all || $lowres || $highres) && die "Cannot use option -m with options -a, -l or -h"
+  echo "Generating new config.yml file with middle resources from config-2_1.yml"
+  patch -o $BASEDIR/config-2_1.yml.MIDRES $BASEDIR/config-2_1.yml $BASEDIR/config-2_1.yml.mid_res.patch
+  circleci config process $BASEDIR/config-2_1.yml.MIDRES > $BASEDIR/config.yml.MIDRES.tmp
+  cat $BASEDIR/license.yml $BASEDIR/config.yml.MIDRES.tmp > $BASEDIR/config.yml
+  rm $BASEDIR/config-2_1.yml.MIDRES $BASEDIR/config.yml.MIDRES.tmp
+
+elif $highres; then
+  ($all || $lowres || $midres) && die "Cannot use option -h with options -a, -l or -m"
+  echo "Generating new config.yml file with high resources from config-2_1.yml"
+  patch -o $BASEDIR/config-2_1.yml.HIGHRES $BASEDIR/config-2_1.yml $BASEDIR/config-2_1.yml.high_res.patch
+  circleci config process $BASEDIR/config-2_1.yml.HIGHRES > $BASEDIR/config.yml.HIGHRES.tmp
+  cat $BASEDIR/license.yml $BASEDIR/config.yml.HIGHRES.tmp > $BASEDIR/config.yml
+  rm $BASEDIR/config-2_1.yml.HIGHRES $BASEDIR/config.yml.HIGHRES.tmp
 
 elif $all; then
-  ($free || $paid || $has_env_vars) && die "Cannot use option -a with options -f, -p or -e"
-  echo "Generating new default config.yml file for free tier and FREE/PAID templates from config_template.yml."
-  echo "Make sure you commit the newly generated config.yml, config.yml.FREE and config.yml.PAID files"
-  echo "after running this command if you want them to persist."
+  ($lowres || $midres || $highres || $has_env_vars) && die "Cannot use option -a with options -l, -m, -h or -e"
+  echo "Generating new config.yml file with low resources and LOWRES/MIDRES/HIGHRES templates from config-2_1.yml"
 
-  # setup config for free tier
-  circleci config process $BASEDIR/config_template.yml > $BASEDIR/config.yml.FREE.tmp
-  cat $BASEDIR/license.yml $BASEDIR/config.yml.FREE.tmp > $BASEDIR/config.yml.FREE
-  rm $BASEDIR/config.yml.FREE.tmp
+  # setup lowres
+  circleci config process $BASEDIR/config-2_1.yml > $BASEDIR/config.yml.LOWRES.tmp
+  cat $BASEDIR/license.yml $BASEDIR/config.yml.LOWRES.tmp > $BASEDIR/config.yml.LOWRES
+  rm $BASEDIR/config.yml.LOWRES.tmp
 
-  # setup config for paid tier
-  patch -o $BASEDIR/config_template.yml.PAID $BASEDIR/config_template.yml $BASEDIR/config_template.yml.PAID.patch
-  circleci config process $BASEDIR/config_template.yml.PAID > $BASEDIR/config.yml.PAID.tmp
-  cat $BASEDIR/license.yml $BASEDIR/config.yml.PAID.tmp > $BASEDIR/config.yml.PAID
-  rm $BASEDIR/config_template.yml.PAID $BASEDIR/config.yml.PAID.tmp
+  # setup midres
+  patch -o $BASEDIR/config-2_1.yml.MIDRES $BASEDIR/config-2_1.yml $BASEDIR/config-2_1.yml.mid_res.patch
+  circleci config process $BASEDIR/config-2_1.yml.MIDRES > $BASEDIR/config.yml.MIDRES.tmp
+  cat $BASEDIR/license.yml $BASEDIR/config.yml.MIDRES.tmp > $BASEDIR/config.yml.MIDRES
+  rm $BASEDIR/config-2_1.yml.MIDRES $BASEDIR/config.yml.MIDRES.tmp
 
-  # copy free tier into config.yml to make sure this gets updated
-  cp $BASEDIR/config.yml.FREE $BASEDIR/config.yml
+  # setup highres
+  patch -o $BASEDIR/config-2_1.yml.HIGHRES $BASEDIR/config-2_1.yml $BASEDIR/config-2_1.yml.high_res.patch
+  circleci config process $BASEDIR/config-2_1.yml.HIGHRES > $BASEDIR/config.yml.HIGHRES.tmp
+  cat $BASEDIR/license.yml $BASEDIR/config.yml.HIGHRES.tmp > $BASEDIR/config.yml.HIGHRES
+  rm $BASEDIR/config-2_1.yml.HIGHRES $BASEDIR/config.yml.HIGHRES.tmp
 
-elif (!($has_env_vars)); then
+  # copy lower into config.yml to make sure this gets updated
+  cp $BASEDIR/config.yml.LOWRES $BASEDIR/config.yml
+
+elif (! ($has_env_vars)); then
   print_help
   exit 0
 fi
 
 # add new or modified tests to the sets of tests to be repeated
-if $detect_changed_tests; then
-  # Sanity check that the referenced branch exists
-  if ! git show ${BASE_BRANCH} -- >&/dev/null; then
-    echo -e "\n\nUnknown base branch: ${BASE_BRANCH}. Unable to detect changed tests.\n"
-    echo    "Please use the '-b' option to choose an existing branch name"
-    echo    "(e.g. origin/${BASE_BRANCH}, apache/${BASE_BRANCH}, etc.)."
-    exit 2
-  fi
-
+if (! ($all)); then
   add_diff_tests ()
   {
     dir="${BASEDIR}/../${2}"
@@ -205,8 +207,7 @@ if $detect_changed_tests; then
            | sed -e "s/\\.java//" \
            | sed -e "s,^${2},," \
            | tr  '/' '.' \
-           | grep ${3} )\
-           || : # avoid execution interruptions due to grep return codes and set -e
+           | grep ${3} )
     for test in $tests; do
       echo "  $test"
       has_env_vars=true
@@ -226,6 +227,7 @@ if $detect_changed_tests; then
   add_diff_tests "REPEATED_UTESTS_LONG" "test/long/" "org.apache.cassandra"
   add_diff_tests "REPEATED_UTESTS_STRESS" "tools/stress/test/unit/" "org.apache.cassandra.stress"
   add_diff_tests "REPEATED_UTESTS_FQLTOOL" "tools/fqltool/test/unit/" "org.apache.cassandra.fqltool"
+  add_diff_tests "REPEATED_SIMULATOR_DTESTS" "test/simulator/test/" "org.apache.cassandra.simulator.test"
   add_diff_tests "REPEATED_JVM_DTESTS" "test/distributed/" "org.apache.cassandra.distributed.test"
   add_diff_tests "REPEATED_JVM_UPGRADE_DTESTS" "test/distributed/" "org.apache.cassandra.distributed.upgrade"
 fi
@@ -255,8 +257,8 @@ delete_job()
     sed -Ei.bak "/^    - ${2}/d" "$1"
   }
   file="$BASEDIR/$1"
-  delete_yaml_block "$file" "${2}:"
-  delete_yaml_block "$file" "start_${2}:"
+  delete_yaml_block "$file" "${2}"
+  delete_yaml_block "$file" "start_${2}"
 }
 
 # Define function to remove any unneeded repeated jobs.
@@ -285,6 +287,10 @@ delete_repeated_jobs()
     delete_job "$1" "j8_utests_fqltool_repeat"
     delete_job "$1" "j11_utests_fqltool_repeat"
   fi
+  if (! (echo "$env_vars" | grep -q "REPEATED_SIMULATOR_DTESTS=")); then
+    delete_job "$1" "j8_simulator_dtests_repeat"
+    delete_job "$1" "j11_simulator_dtests_repeat"
+  fi
   if (! (echo "$env_vars" | grep -q "REPEATED_JVM_DTESTS=")); then
     delete_job "$1" "j8_jvm_dtests_repeat"
     delete_job "$1" "j8_jvm_dtests_vnode_repeat"
@@ -320,6 +326,7 @@ delete_repeated_jobs()
 
 delete_repeated_jobs "config.yml"
 if $all; then
-  delete_repeated_jobs "config.yml.FREE"
-  delete_repeated_jobs "config.yml.PAID"
+  delete_repeated_jobs "config.yml.LOWRES"
+  delete_repeated_jobs "config.yml.MIDRES"
+  delete_repeated_jobs "config.yml.HIGHRES"
 fi
