@@ -28,13 +28,12 @@ import org.slf4j.LoggerFactory;
 
 import org.apache.cassandra.config.Config;
 import org.apache.cassandra.db.*;
-import org.apache.cassandra.db.partitions.Partition;
+import org.apache.cassandra.db.memtable.Memtable;
 import org.apache.cassandra.db.rows.UnfilteredRowIterator;
 import org.apache.cassandra.io.sstable.format.SSTableReader;
 import org.apache.cassandra.io.util.FileDataInput;
 import org.apache.cassandra.io.util.FileUtils;
 import org.apache.cassandra.schema.CompactionParams.TombstoneOption;
-import org.apache.cassandra.utils.AlwaysPresentFilter;
 import org.apache.cassandra.utils.OverlapIterator;
 import org.apache.cassandra.utils.concurrent.Refs;
 
@@ -73,8 +72,6 @@ public class CompactionController extends AbstractCompactionController
 
     public CompactionController(ColumnFamilyStore cfs, Set<SSTableReader> compacting, int gcBefore, RateLimiter limiter, TombstoneOption tombstoneOption)
     {
-        //When making changes to the method, be aware that some of the state of the controller may still be uninitialized
-        //(e.g. TWCS sets up the value of ignoreOverlaps() after this completes)
         super(cfs, gcBefore, tombstoneOption);
         this.compacting = compacting;
         this.limiter = limiter;
@@ -96,6 +93,12 @@ public class CompactionController extends AbstractCompactionController
             return;
         }
 
+        if (ignoreOverlaps())
+        {
+            logger.debug("not refreshing overlaps - running with ignoreOverlaps activated");
+            return;
+        }
+
         if (cfs.getNeverPurgeTombstones())
         {
             logger.debug("not refreshing overlaps for {}.{} - neverPurgeTombstones is enabled", cfs.keyspace.getName(), cfs.getTableName());
@@ -112,7 +115,7 @@ public class CompactionController extends AbstractCompactionController
         }
     }
 
-    void refreshOverlaps()
+    private void refreshOverlaps()
     {
         if (NEVER_PURGE_TOMBSTONES || cfs.getNeverPurgeTombstones())
             return;
@@ -120,7 +123,7 @@ public class CompactionController extends AbstractCompactionController
         if (this.overlappingSSTables != null)
             close();
 
-        if (compacting == null)
+        if (compacting == null || ignoreOverlaps())
             overlappingSSTables = Refs.tryRef(Collections.<SSTableReader>emptyList());
         else
             overlappingSSTables = cfs.getAndReferenceOverlappingLiveSSTables(compacting);
@@ -198,10 +201,7 @@ public class CompactionController extends AbstractCompactionController
         }
 
         for (Memtable memtable : cfStore.getTracker().getView().getAllMemtables())
-        {
-            if (memtable.getMinTimestamp() != Memtable.NO_MIN_TIMESTAMP)
-                minTimestamp = Math.min(minTimestamp, memtable.getMinTimestamp());
-        }
+            minTimestamp = Math.min(minTimestamp, memtable.getMinTimestamp());
 
         // At this point, minTimestamp denotes the lowest timestamp of any relevant
         // SSTable or Memtable that contains a constructive value. candidates contains all the
@@ -254,10 +254,7 @@ public class CompactionController extends AbstractCompactionController
 
         for (SSTableReader sstable: filteredSSTables)
         {
-            // if we don't have bloom filter(bf_fp_chance=1.0 or filter file is missing),
-            // we check index file instead.
-            if (sstable.getBloomFilter() instanceof AlwaysPresentFilter && sstable.getPosition(key, SSTableReader.Operator.EQ, false) != null
-                || sstable.getBloomFilter().isPresent(key))
+            if (sstable.maybePresent(key))
             {
                 minTimestampSeen = Math.min(minTimestampSeen, sstable.getMinTimestamp());
                 hasTimestamp = true;
@@ -266,14 +263,10 @@ public class CompactionController extends AbstractCompactionController
 
         for (Memtable memtable : memtables)
         {
-            if (memtable.getMinTimestamp() != Memtable.NO_MIN_TIMESTAMP)
+            if (memtable.rowIterator(key) != null)
             {
-                Partition partition = memtable.getPartition(key);
-                if (partition != null)
-                {
-                    minTimestampSeen = Math.min(minTimestampSeen, partition.stats().minTimestamp);
-                    hasTimestamp = true;
-                }
+                minTimestampSeen = Math.min(minTimestampSeen, memtable.getMinTimestamp());
+                hasTimestamp = true;
             }
         }
 
@@ -340,8 +333,6 @@ public class CompactionController extends AbstractCompactionController
      * This strategy can retain for a long time a lot of sstables on disk (see CASSANDRA-13418) so this option
      * control whether or not this check should be ignored.
      *
-     * Do NOT call this method in the CompactionController constructor
-     *
      * @return false by default
      */
     protected boolean ignoreOverlaps()
