@@ -56,7 +56,7 @@ public class LeveledCompactionTask extends CompactionTask
     @Override
     protected boolean partialCompactionsAcceptable()
     {
-        return level == 0;
+        throw new UnsupportedOperationException("This is now handled in reduceScopeForLimitedSpace");
     }
 
     protected int getLevel()
