@@ -17,7 +17,6 @@
  */
 package org.apache.cassandra.io.sstable;
 
-import java.io.File;
 import java.util.Collections;
 import java.util.List;
 import java.util.Objects;
@@ -25,6 +24,7 @@ import java.util.concurrent.CountDownLatch;
 
 import com.google.common.io.Files;
 
+import org.apache.cassandra.io.util.File;
 import org.junit.After;
 import org.junit.Before;
 import org.junit.BeforeClass;
@@ -33,7 +33,6 @@ import org.junit.Test;
 import org.apache.cassandra.SchemaLoader;
 import org.apache.cassandra.Util;
 import org.apache.cassandra.locator.Replica;
-import org.apache.cassandra.schema.TableId;
 import org.apache.cassandra.schema.TableMetadata;
 import org.apache.cassandra.schema.TableMetadataRef;
 import org.apache.cassandra.schema.Schema;
@@ -61,7 +60,6 @@ public class SSTableLoaderTest
     public static final String CF_STANDARD1 = "Standard1";
     public static final String CF_STANDARD2 = "Standard2";
     public static final String CF_BACKUPS = Directories.BACKUPS_SUBDIR;
-    public static final String CF_SNAPSHOTS = Directories.SNAPSHOT_SUBDIR;
 
     private static final String schema = "CREATE TABLE %s.%s (key ascii, name ascii, val ascii, val1 ascii, PRIMARY KEY (key, name))";
     private static final String query = "INSERT INTO %s.%s (key, name, val) VALUES (?, ?, ?)";
@@ -76,8 +74,7 @@ public class SSTableLoaderTest
                                     KeyspaceParams.simple(1),
                                     SchemaLoader.standardCFMD(KEYSPACE1, CF_STANDARD1),
                                     SchemaLoader.standardCFMD(KEYSPACE1, CF_STANDARD2),
-                                    SchemaLoader.standardCFMD(KEYSPACE1, CF_BACKUPS),
-                                    SchemaLoader.standardCFMD(KEYSPACE1, CF_SNAPSHOTS));
+                                    SchemaLoader.standardCFMD(KEYSPACE1, CF_BACKUPS));
 
         SchemaLoader.createKeyspace(KEYSPACE2,
                 KeyspaceParams.simple(1),
@@ -90,7 +87,7 @@ public class SSTableLoaderTest
     @Before
     public void setup() throws Exception
     {
-        tmpdir = Files.createTempDir();
+        tmpdir = new File(Files.createTempDir());
     }
 
     @After
@@ -100,7 +97,6 @@ public class SSTableLoaderTest
             FileUtils.deleteRecursive(tmpdir);
         } catch (FSWriteError e) {
             /*
-              Windows does not allow a mapped file to be deleted, so we probably forgot to clean the buffers somewhere.
               We force a GC here to force buffer deallocation, and then try deleting the directory again.
               For more information, see: http://bugs.java.com/bugdatabase/view_bug.do?bug_id=4715154
               If this is not the problem, the exception will be rethrown anyway.
@@ -143,7 +139,7 @@ public class SSTableLoaderTest
         }
 
         ColumnFamilyStore cfs = Keyspace.open(KEYSPACE1).getColumnFamilyStore(CF_STANDARD1);
-        cfs.forceBlockingFlush(); // wait for sstables to be on disk else we won't be able to stream them
+        Util.flush(cfs); // wait for sstables to be on disk else we won't be able to stream them
 
         final CountDownLatch latch = new CountDownLatch(1);
         SSTableLoader loader = new SSTableLoader(dataDir, new TestClient(), new OutputHandler.SystemOutput(false, false));
@@ -175,7 +171,7 @@ public class SSTableLoaderTest
                                                   .withBufferSizeInMB(1)
                                                   .build();
 
-        int NB_PARTITIONS = 5000; // Enough to write >1MB and get at least one completed sstable before we've closed the writer
+        int NB_PARTITIONS = 5000; // Enough to write >1MiB and get at least one completed sstable before we've closed the writer
 
         for (int i = 0; i < NB_PARTITIONS; i++)
         {
@@ -184,10 +180,10 @@ public class SSTableLoaderTest
         }
 
         ColumnFamilyStore cfs = Keyspace.open(KEYSPACE1).getColumnFamilyStore(CF_STANDARD2);
-        cfs.forceBlockingFlush(); // wait for sstables to be on disk else we won't be able to stream them
+        Util.flush(cfs); // wait for sstables to be on disk else we won't be able to stream them
 
         //make sure we have some tables...
-        assertTrue(Objects.requireNonNull(dataDir.listFiles()).length > 0);
+        assertTrue(Objects.requireNonNull(dataDir.tryList()).length > 0);
 
         final CountDownLatch latch = new CountDownLatch(2);
         //writer is still open so loader should not load anything
@@ -215,8 +211,8 @@ public class SSTableLoaderTest
     @Test
     public void testLoadingSSTableToDifferentKeyspace() throws Exception
     {
-        File dataDir = new File(tmpdir.getAbsolutePath() + File.separator + KEYSPACE1 + File.separator + CF_STANDARD1);
-        assert dataDir.mkdirs();
+        File dataDir = new File(tmpdir.absolutePath() + File.pathSeparator() + KEYSPACE1 + File.pathSeparator() + CF_STANDARD1);
+        assert dataDir.tryCreateDirectories();
         TableMetadata metadata = Schema.instance.getTableMetadata(KEYSPACE1, CF_STANDARD1);
 
         String schema = "CREATE TABLE %s.%s (key ascii, name ascii, val ascii, val1 ascii, PRIMARY KEY (key, name))";
@@ -232,14 +228,14 @@ public class SSTableLoaderTest
         }
 
         ColumnFamilyStore cfs = Keyspace.open(KEYSPACE1).getColumnFamilyStore(CF_STANDARD1);
-        cfs.forceBlockingFlush(); // wait for sstables to be on disk else we won't be able to stream them
+        Util.flush(cfs); // wait for sstables to be on disk else we won't be able to stream them
 
         final CountDownLatch latch = new CountDownLatch(1);
         SSTableLoader loader = new SSTableLoader(dataDir, new TestClient(), new OutputHandler.SystemOutput(false, false), 1, KEYSPACE2);
         loader.stream(Collections.emptySet(), completionStreamListener(latch)).get();
 
         cfs = Keyspace.open(KEYSPACE2).getColumnFamilyStore(CF_STANDARD1);
-        cfs.forceBlockingFlush();
+        Util.flush(cfs);
 
         List<FilteredPartition> partitions = Util.getAll(Util.cmd(cfs).build());
 
@@ -258,48 +254,20 @@ public class SSTableLoaderTest
     @Test
     public void testLoadingBackupsTable() throws Exception
     {
-        testLoadingTable(CF_BACKUPS);
-    }
-
-    @Test
-    public void testLoadingSnapshotsTable() throws Exception
-    {
-        testLoadingTable(CF_SNAPSHOTS);
-    }
-
-    private void testLoadingTable(String tableName) throws Exception
-    {
-        testLoadingTable(CF_BACKUPS, false);
-    }
-
-    @Test
-    public void testLoadingLegacyBackupsTable() throws Exception
-    {
-        testLoadingTable(CF_BACKUPS, true);
-    }
-
-    @Test
-    public void testLoadingLegacySnapshotsTable() throws Exception
-    {
-        testLoadingTable(CF_SNAPSHOTS, true);
-    }
-
-    private void testLoadingTable(String tableName, boolean isLegacyTable) throws Exception
-    {
-        File dataDir = dataDir(tableName, isLegacyTable);
-        TableMetadata metadata = Schema.instance.getTableMetadata(KEYSPACE1, tableName);
+        File dataDir = dataDir(CF_BACKUPS);
+        TableMetadata metadata = Schema.instance.getTableMetadata(KEYSPACE1, CF_BACKUPS);
 
         try (CQLSSTableWriter writer = CQLSSTableWriter.builder()
                                                        .inDirectory(dataDir)
-                                                       .forTable(String.format(schema, KEYSPACE1, tableName))
-                                                       .using(String.format(query, KEYSPACE1, tableName))
+                                                       .forTable(String.format(schema, KEYSPACE1, CF_BACKUPS))
+                                                       .using(String.format(query, KEYSPACE1, CF_BACKUPS))
                                                        .build())
         {
             writer.addRow("key", "col1", "100");
         }
 
-        ColumnFamilyStore cfs = Keyspace.open(KEYSPACE1).getColumnFamilyStore(tableName);
-        cfs.forceBlockingFlush(); // wait for sstables to be on disk else we won't be able to stream them
+        ColumnFamilyStore cfs = Keyspace.open(KEYSPACE1).getColumnFamilyStore(CF_BACKUPS);
+        Util.flush(cfs); // wait for sstables to be on disk else we won't be able to stream them
 
         final CountDownLatch latch = new CountDownLatch(1);
         SSTableLoader loader = new SSTableLoader(dataDir, new TestClient(), new OutputHandler.SystemOutput(false, false));
@@ -321,17 +289,10 @@ public class SSTableLoaderTest
 
     private File dataDir(String cf)
     {
-        return dataDir(cf, false);
-    }
-
-    private File dataDir(String cf, boolean isLegacyTable)
-    {
-        // Add -{tableUuid} suffix to table dir if not a legacy table
-        File dataDir = new File(tmpdir.getAbsolutePath() + File.separator + SSTableLoaderTest.KEYSPACE1 + File.separator + cf
-                                + (isLegacyTable ? "" : String.format("-%s", TableId.generate().toHexString())));
-        assert dataDir.mkdirs();
+        File dataDir = new File(tmpdir.absolutePath() + File.pathSeparator() + SSTableLoaderTest.KEYSPACE1 + File.pathSeparator() + cf);
+        assert dataDir.tryCreateDirectories();
         //make sure we have no tables...
-        assertEquals(Objects.requireNonNull(dataDir.listFiles()).length, 0);
+        assertEquals(Objects.requireNonNull(dataDir.tryList()).length, 0);
         return dataDir;
     }
 
