@@ -18,8 +18,8 @@
  */
 package org.apache.cassandra.utils;
 
-import java.io.File;
 
+import org.apache.cassandra.io.util.File;
 import org.junit.Assert;
 import org.junit.Test;
 
@@ -32,7 +32,7 @@ public class NativeLibraryTest
     {
         File file = FileUtils.createDeletableTempFile("testSkipCache", "1");
 
-        NativeLibrary.trySkipCache(file.getPath(), 0, 0);
+        NativeLibrary.trySkipCache(file.path(), 0, 0);
     }
 
     @Test
